#!/usr/local/bin/perl 
# ------------------
# e-merge 1.2.3 -  5/14/99
# http://www.usfca.edu/turner/e-merge/
#
# Benjamin Turner <bjturner@bigfoot.com>
# http://www.usfca.edu/turner/
#
# (c) Copyright 1995-1999, Benjamin Turner
#
# =====================================================================
#
#   e-merge takes the input from an HTML form, does a 'mail merge' of
# the input into a specified merge file, and sends the mail merge as
# e-mail.
#   Please see <http://www.usfca.edu/turner/e-merge> for directions 
# on how to use e-merge.
#
# Changes in 1.2.3:
#
#   *   E-merge can now use regexp-special character as a LOGDELIMITER
#       such as | or * or . or ?
#
# Changes in 1.2.2:
#
#   *   E-merge wasn't properly handling web servers on ports other 
#       than 80 (the standard http port), since it was omitting the
#       port number when it printed the BASE HREF URL in the web
#       page sent back to the browser. This should now be fixed.
#
# Changes in 1.2.1:
#
#   *   Well, you fix one bug, you create another. In adding the 
#       ability to use fields in the log that weren't in the merge 
#       file, I broke e-merge for those *not* using a log file (you'd 
#       get a message like "Can't open log file: No such file or 
#       directory"). Fixed by only looking for fields in a log file
#       if the logging option has been turned on. A similar problem
#       with display files was also fixed.
#
# Changes in 1.2:
# 
#   *   Looks like the log file had the same problem with using fields
#       that weren't in the merge file. Also added the ability to use
#       environment variables in the log file. While I was mucking 
#       around, I added the option to use alternate logfile field
#       delimiters; however, it's not something you can set in the
#       ACTION line, so you've got to be able to alter the e-merge
#       script itself to do it.
# 
# Changes in 1.1.4:
#
#   *   Previously, e-merge could not substitute field values in the 
#       display file unless those fields were used in the mail merge 
#       file. This has been fixed.
#
# Changes in 1.1.3:
#
#   *   Hey! e-merge now actually logs like it's supposed to! Somehow,
#       a stray 'c' wandered into the function that stripped out CR/LF
#       and caused it to strip out everything *except* CR/LF. Duh!
#
# Changes in 1.1.2:
#
#   *   Fixed a problem with using relative URLs in the display file
#       that was causing data-less mail messages to be sent.
#
# Changes in 1.1.1:
#
#   *   Fixed some spurious HTML preceding the output of the display
#       file which made everything part of an <H2> tag.
#
# Changes in 1.1:
#
#   *   e-merge can now use a 'display' file to control the response
#       page separately from the normal mail 'merge' file. 
#
#   *   Fixed a bug with the log file code which caused the pathname
#       for the log file to get screwed up in some cases. Duh...
#
# For a full version history, see 
#  http://www.usfca.edu/usf/turner/e-merge/versions.html
#
# ----
# Permission is granted to use and modify this software to your heart's
# content, but don't charge for anything based on it. I'd love to be
# kept informed of what you've done with it or what you'd like to see
# it do, but I know that most of you won't bother to tell me. :)
# And remember, if this software fails or screws up or hurts anything,
# the risk is on your head, not mine -- I take no responsibility for 
# how this software works for you. Having said that... Enjoy!
#
#######################################################################
# Libraries #
#############
require 5.00;                           # Perl 5 references used

#######################################################################
# Global Variables #
####################

### MAILPROGRAM ###
# you may need to change this to reflect the real location of sendmail
# or to specify a different mail program to run (it must be able to
# read its headers from the data piped to it).

$mailprogram = "/usr/lib/sendmail -t";

### XSENDER ###
# This identifies e-merge as the sender of the mail and has a pointer
# to the e-merge page for those interested in e-merge. It get spliced
# into the merge file near the top.

$xsender = "e-merge <URL:http://www.usfca.edu/usf/turner/e-merge/>";

### LOGDELIMITER ###
# This is the delimiter used for the log file. About the only other
# likely-to-be-useful value is "," but you can put whatever you want.
# Just watch out that no one sticks this character in the data.

$LOGDELIMITER = "\t";

#######################################################################
# Main Program #
################

print "Content-Type: text/html\n\n";   # print out our header
%FORM = &RetrieveData;                 # Get user's data from form

# parse info about merge, display, and log files, and testing status
( $mergefile, $displayfile, $displayURL, $TESTING, $LOGGING ) 
   = &GetOptions ( $ENV{'PATH_TRANSLATED'} );

# read the merge file into an array for processing.
open ( MERGE, $mergefile )
   or &FatalError ("Can't open merge file $mergefile: $!");
@MERGE = <MERGE>;                  # slurp the file for later scanning
close MERGE;
splice ( @MERGE, 3, 0, "X-Sender: " . $xsender . "\n" );

# Find all fields, and look to see if any required fields are empty
%MERGE = &FindMergeFields ( \%FORM, @MERGE );

if ( $displayfile )
{
   # read in display file, too
   open( DISPLAY, $displayfile )
      or &FatalError( "Can't open display file $displayfile: $!" );
   @DISPLAY = <DISPLAY>;
   close DISPLAY;

   %MERGE = ( %MERGE, &FindMergeFields( \%FORM, @DISPLAY ) );
}

%MERGE = ( %MERGE, &FindMergeFieldsInLog( \%FORM, $LOGGING ) ) if $LOGGING;

@empty = &FindEmptyFields ( \%FORM, &RequiredFields ( %MERGE ) );
&EmptyError ( @empty ) if ( @empty );  # chastise user for blanks

# Do field-value substitution on merge array (it's really that easy...)
foreach (@MERGE) { s|\[(.*?)\]|$MERGE{$1}|g; }

# Send off the merge array as mail (it'd better have valid headers)
unless ( $TESTING ) {
  &SendArrayAsMail ( @MERGE );
  &LogInput ( $LOGGING, %MERGE ) if $LOGGING;
}

# if user requests a display file, merge that; otherwise, show mail
if ($displayfile)
{
  # Do field-value substitution for display file
  foreach (@DISPLAY) { s|\[(.*?)\]|$MERGE{$1}|g; }
  
  &ShowDisplay ( $displayURL, @DISPLAY );
}
else { &ShowMail ( @MERGE ); }


#######################################################################
# Function Definitions #
########################
#
#######################################################################
# Get merge file path and testing status ( &test at end of Path )
sub GetOptions
{
  my ( $mergefile ) = @_;
  my ( $display, $testing, $logfile );

  $testing = ( $mergefile =~ s/&test// );   # check for test mode

  # now check to see if we should use a display file (&display present?)
  $display = "display$1" if ( $mergefile =~ s/&display(=[^&]+)?// );
  if ( $display ) { 
	# if no display file was explicitly named, make it mergefile.display
	unless ( $display =~ s|display=|| )
    {
      ( $display = $mergefile ) =~ s#\.[^/]*?$|$#.display#;
      $displayURL = $ENV{'PATH_INFO'};
      $displayURL = s|&.*||;
    }
    else
    {
      $displayURL = $display;           # save URL form for BASE HREF

      # if $display is in ~user form, use PATH_INFO to get partial path
      unless ( $display =~ s|/?~[^/]+|&GetPathToUserDir|e ) 

      # otherwise, just tack DOCUMENT_ROOT onto the front
      { $display = $ENV{'DOCUMENT_ROOT'} . $display; }
    }

    # Make extra-sure there are no .. refs in the displayfile path (the 
    # server usually takes care of this
    $display =~ s|\.\./||g;
  }

  # now check to see if we should log (&log present?) and where
  $logfile = "log$1" if ( $mergefile =~ s/&log(=[^&]+)?// );
  if ( $logfile ) { 
	# if no log file was explicitly directed, make it mergefile.log
	unless ( $logfile =~ s|log=|| )
      { ( $logfile = $mergefile ) =~ s#\.[^/]*?$|$#.log#; }
    else
    {
      # if $logfile is in ~user form, use PATH_INFO to get partial path
      unless ( $logfile =~ s|/?~[^/]+|&GetPathToUserDir|e ) 

      # otherwise, just tack DOCUMENT_ROOT onto the front
      { $logfile = $ENV{'DOCUMENT_ROOT'} . $logfile; }
    }

    # Make extra-sure there are no .. refs in the logfile path (the 
    # server usually takes care of this
    $logfile =~ s|\.\./||g;
  }

  $mergefile =~ s/&.*//g;                   # toss any other options

  return( $mergefile, $display, $displayURL, $testing, $logfile );
}

#######################################################################
# Makes a hash of all fields in the merge file and their corresponding
# values from the user data from the html form.
sub FindMergeFields
{
  my ( $FORM, @MERGE ) = @_;
  my ( %MERGE, @fields );

  foreach ( @MERGE ) {
	@fields = m/\[.*?\]/g;                # get fields on this line
	foreach $field ( @fields ) {
      $field =~ tr|[]||d;                 # chop off the [] brackets
	  if ( $field =~ m/^\$(.+)/ )         # if the first letter is $
		{ $MERGE{$field} = $ENV{$1}; }    # then use env. var. w/o $
	  else {                              # otherwise, use field value
		$MERGE{$field} = $$FORM{$field};
		if ( $MERGE{$field} =~ m/\0/ ) {  # if there are > 1 values
		  if ( m/^(\s*)\[$field\]/ ) {    # check if the field is pre-
		    $blankspace = $1;             # ceded only by white space
			$MERGE{$field} =~ s|\0|\n$blankspace|g; # if so, put it in
		  }                               # + \n before each value
		  else { $MERGE{$field} =~ s|\0|, |g; } # else, comma & space
        }
      }
	}
  }
  return( %MERGE );
}

#######################################################################
# Makes a hash of all fields in the log file and their corresponding
# values from the user data from the html form + environment variables.
sub FindMergeFieldsInLog
{
  my ( $FORM, $logfile ) = @_;
  my ( %MERGE, @fieldlist, $field, $formatline );

  if ( -e $logfile )
  {
	open (LOG, "$logfile") or &FatalError("Can't open log file: $!");

    # Read first line to decide in what order to write field data
	chomp ( $formatline = <LOG> );
    @fieldlist = split ( /\Q$LOGDELIMITER\E/, $formatline );
    close LOG;
  }

  # if there isn't yet a log file, create one and write format line
  else
  {
	open (LOG, ">$logfile") or &FatalError ("Can't write log file: $!");

    # We'll write the format line to put the fields in alphabetic order
	@fieldlist = sort ( keys %FORM );
    print LOG join ( $LOGDELIMITER, @fieldlist ), "\n";
    close LOG;
  }

  # add any fields in log to the Merge list
  foreach $field ( @fieldlist )         # for each field in log
  {
    if ( $field =~ m/^\$(.+)/ )         # if name starts with $
      { $MERGE{$field} = $ENV{$1}; }    # substitute with env. var.
    else                                # otherwise user the
      { $MERGE{$field} = $FORM->{$field}; } # field value
  }

  return( %MERGE );
}

#######################################################################
# Build a list of fields from merge file which are marked 'required-'
sub RequiredFields
{
  my ( %MERGE ) = @_;
  my ( @required );

  foreach ( sort ( keys %MERGE ) )
	{ push ( @required, $_ ) if ( m/^required-/ ); }
  return( @required );
}

#######################################################################
# Sends the Array as a mail message. Doesn't check mail headers!
sub SendArrayAsMail
{
  my ( @MAIL ) = @_;

  open(MAIL, "|$mailprogram") or &FatalError ("Can't run sendmail: $!");
  foreach (@MAIL) { print MAIL; }
  close MAIL;
}

#######################################################################
# Returns the path to a user's web directory if the path info was a
# ~user directory.
sub GetPathToUserDir
{
  my ( $userpath, $afterusername );

  $userpath = $ENV{'PATH_TRANSLATED'};

  # Strip out /~username part from PATH_INFO
  ( $afterusername = $ENV{'PATH_INFO'} ) =~ s|^/~[^/]+||;

  # Then see what that part got translated to in PATH_TRANSLATED
  $userpath =~ s|(.*)$afterusername|$1|;
  return( $userpath );
}

#######################################################################
# Record the fields to a tab-delimited file. Reads the first line of
# the file to determine the order the fields should be placed in.
sub LogInput
{
  my ( $logfile, %input ) = @_;
  my ( $formatline, $logline );

  # if there's already a log file, just read format before appending 
  if ( -e $logfile )
  {
	open (LOG, $logfile) or &FatalError("Can't open log file: $!");

    # Read first line to decide in what order to write field data
	chomp ( $formatline = <LOG> );
    @fieldlist = split ( /\Q$LOGDELIMITER\E/, $formatline );
	open (LOG, ">>$logfile") or &FatalError("Can't write log file: $!");
  }
  # if there isn't yet a log file, create one and write format line
  else {
	open (LOG, ">$logfile") or &FatalError ("Can't write log file: $!");

    # We'll write the format line to put the fields in alphabetic order
	@fieldlist = sort ( keys %input );
    print LOG join ( $LOGDELIMITER, @fieldlist ), "\n";
  }

  # look up values for fields and write them to the logfile.
  $logline = join ( $LOGDELIMITER, &ValuesForKeys ( \%input, @fieldlist ) );
  $logline =~ tr/\015\012/ /s;             # change CR & LF to space
  print LOG "$logline\n";
  
  close LOG;
}

#######################################################################
# Returns a list of of the values for the keys of the input array which
# are listed in the list following the ref to the input array.
sub ValuesForKeys
{
  my ( $hash, @keys ) = @_;
  my ( @values, $value );

  foreach $key (@keys) {
	$value = $$hash{$key};                 # look up value for key
    $value =~ s/\0/, /g;                   # replace nulls with ', '
	push ( @values, $value );              # pop value onto return list
  }

  return( @values );
}

#######################################################################
# Shows the user the merged display file
sub ShowDisplay
{
	my ( $displayURL, @MERGE ) = @_;
	my ( $serverAddress ) = $ENV{'SERVER_NAME'};
  
	if ( $ENV{'SERVER_PORT'} != 80 )
		{ $serverAddress .= ":$ENV{'SERVER_PORT'}"; }

	print "<BASE HREF=\"http://$serverAddress$displayURL\">\n";
	
	if ( $TESTING )
	{
		print "<TITLE>Test Results</TITLE>\n<H2>The following page ",
			"would have been displayed:</H2>\n<HR>\n";
	}
	foreach ( @MERGE ) { print; }          # print each line of MERGE
}

#######################################################################
# Merely shows the user what the mail looked like when it was sent
sub ShowMail
{
  my ( @MERGE ) = @_;

  if ( $TESTING ) {
	print "<TITLE>Test Results</TITLE>\n<H2>The following message ",
		  "would have been sent:</H2>\n<HR>\n<PRE>";
  }
  else {
    print "<TITLE>Mail Sent</TITLE>\n<H2>The following message was ",
	    "sent:</H2>\n<HR>\n<PRE>";
  }
  foreach ( @MERGE ) { print; }          # print each line of MERGE
  print "</PRE>\n<HR>\n";
}

#######################################################################
# Displays error message with a list of which fields must yet be filled
sub EmptyError
{
  my ( @emptyfields ) = @_;

  print "<TITLE>Error</TITLE>\n<H1>Some Required Fields were left ",
		"Empty</H1>\nYou did not fill in values for the following ",
		"required fields:<P>\n<UL>\n";
  # now show user which field names they must enter
  foreach ( @emptyfields ) { print "<LI>$_\n"; }
  exit;
}

#######################################################################
# Print the error message and exit the program
sub FatalError
{
  my ( $error ) = @_;
  print "<TITLE>Error</TITLE>\n<H3>$error</H3>\n";
  exit;
}

#######################################################################
# Reads the data passed to this script from either GET or POST methods
sub RetrieveData
{
  my ( $data, @data, %data, $name, $value );

  # get our data from the appropriate place, depending on METHOD
  if ( $ENV{'REQUEST_METHOD'} eq "GET" )
        { $data = $ENV{'QUERY_STRING'}; }
  else
        { read ( STDIN, $data, $ENV{'CONTENT_LENGTH'} ); }

  # break into separate chunks for each name=value pair
  @data = split ( /&/, $data );

  foreach $data ( @data ) {
    ($name, $value) = split(/=/, $data);

    # Decode the URL-encoding
    $value =~ tr/+/ /;
    $value =~ s/%([A-F0-9][A-F0-9])/pack("C", hex($1))/eig;
        
    # if there are multiple values for the field, splice them together
    # with null characters
    if ( $data{$name} ) { $data{$name} .= "\0"; }
    $data{$name} .= $value;
  }

  return( %data );
}

#######################################################################
# Checks to ensure that the listed keys are in the associative array
# reference passed in in the first argument.
sub FindEmptyFields
{
  my ( $array, @fields ) = @_;
  my ( @empties, $key );

  foreach $key ( @fields )
    { push ( @empties, $key ) unless $$array{$key}; }

  return( @empties );
}
#######################################################################
